


"""
def pretty_print_epica_plan(epicla_plan: List[Union[Instantiated_Action,
                                                    Tuple[Instantiated_Action,
                                                          List[Instantiated_Action]]]]) -> None:
    for action in epicla_plan:
            print(action.name)

"""
